@@ -10,7 +10,7 @@ import threading
 from itertools import count
 from typing import Dict, List, Optional, Set, Text, Tuple, Union, cast
 
-import netifaces
+import ifaddr
 
 from . import mdns, stun, turn
 from .candidate import Candidate, candidate_foundation, candidate_priority
@@ -30,16 +30,41 @@ protocol_id = count()
 _mdns = threading.local()
 
 
-async def get_or_create_mdns_protocol() -> mdns.MDnsProtocol:
+class TransportPolicy(enum.Enum):
+    ALL = 0
+    """
+    All ICE candidates will be considered.
+    """
+
+    RELAY = 1
+    """
+    Only ICE candidates whose IP addresses are being relayed,
+    such as those being passed through a STUN or TURN server,
+    will be considered.
+    """
+
+
+async def get_or_create_mdns_protocol(subscriber: object) -> mdns.MDnsProtocol:
     if not hasattr(_mdns, "lock"):
         _mdns.lock = asyncio.Lock()
         _mdns.protocol = None
+        _mdns.subscribers = set()
     async with _mdns.lock:
         if _mdns.protocol is None:
             _mdns.protocol = await mdns.create_mdns_protocol()
+        _mdns.subscribers.add(subscriber)
     return _mdns.protocol
 
 
+async def unref_mdns_protocol(subscriber: object) -> None:
+    if hasattr(_mdns, "lock"):
+        async with _mdns.lock:
+            _mdns.subscribers.discard(subscriber)
+            if _mdns.protocol and not _mdns.subscribers:
+                await _mdns.protocol.close()
+                _mdns.protocol = None
+
+
 def candidate_pair_priority(
     local: Candidate, remote: Candidate, ice_controlling: bool
 ) -> int:
@@ -56,14 +81,12 @@ def get_host_addresses(use_ipv4: bool, use_ipv6: bool) -> List[str]:
     Get local IP addresses.
     """
     addresses = []
-    for interface in netifaces.interfaces():
-        ifaddresses = netifaces.ifaddresses(interface)
-        for address in ifaddresses.get(socket.AF_INET, []):
-            if use_ipv4 and address["addr"] != "127.0.0.1":
-                addresses.append(address["addr"])
-        for address in ifaddresses.get(socket.AF_INET6, []):
-            if use_ipv6 and address["addr"] != "::1" and "%" not in address["addr"]:
-                addresses.append(address["addr"])
+    for adapter in ifaddr.get_adapters():
+        for ip in adapter.ips:
+            if isinstance(ip.ip, str) and use_ipv4 and ip.ip != "127.0.0.1":
+                addresses.append(ip.ip)
+            elif use_ipv6 and ip.ip[0] != "::1" and ip.ip[2] == 0:
+                addresses.append(ip.ip[0])
     return addresses
 
 
@@ -125,7 +148,7 @@ def validate_remote_candidate(candidate: Candidate) -> Candidate:
 
 class CandidatePair:
     def __init__(self, protocol, remote_candidate: Candidate) -> None:
-        self.handle: Optional[asyncio.Future[None]] = None
+        self.task: Optional[asyncio.Task] = None
         self.nominated = False
         self.protocol = protocol
         self.remote_candidate = remote_candidate
@@ -170,8 +193,9 @@ class StunProtocol(asyncio.DatagramProtocol):
 
     def connection_lost(self, exc: Exception) -> None:
         self.__log_debug("connection_lost(%s)", exc)
-        self.receiver.data_received(None, self.local_candidate.component)
-        self.__closed.set_result(True)
+        if not self.__closed.done():
+            self.receiver.data_received(None, None)
+            self.__closed.set_result(True)
 
     def connection_made(self, transport) -> None:
         self.__log_debug("connection_made(%s)", transport)
@@ -270,6 +294,7 @@ class Connection:
     :param turn_transport: The transport for TURN server, `"udp"` or `"tcp"`.
     :param use_ipv4: Whether to use IPv4 candidates.
     :param use_ipv6: Whether to use IPv6 candidates.
+    :param transport_policy: Transport policy.
     """
 
     def __init__(
@@ -284,6 +309,7 @@ class Connection:
         turn_transport: str = "udp",
         use_ipv4: bool = True,
         use_ipv6: bool = True,
+        transport_policy: TransportPolicy = TransportPolicy.ALL,
     ) -> None:
         self.ice_controlling = ice_controlling
         #: Local username, automatically set to a random value.
@@ -313,6 +339,7 @@ class Connection:
         self._early_checks: List[
             Tuple[stun.Message, Tuple[str, int], StunProtocol]
         ] = []
+        self._early_checks_done = False
         self._event_waiter: Optional[asyncio.Future[ConnectionEvent]] = None
         self._id = next(connection_id)
         self._local_candidates: List[Candidate] = []
@@ -323,12 +350,23 @@ class Connection:
         self._protocols: List[StunProtocol] = []
         self._remote_candidates: List[Candidate] = []
         self._remote_candidates_end = False
-        self._query_consent_handle: Optional[asyncio.Future[None]] = None
+        self._query_consent_task: Optional[asyncio.Task] = None
         self._queue: asyncio.Queue = asyncio.Queue()
         self._tie_breaker = secrets.randbits(64)
         self._use_ipv4 = use_ipv4
         self._use_ipv6 = use_ipv6
 
+        if (
+            stun_server is None
+            and turn_server is None
+            and transport_policy == TransportPolicy.RELAY
+        ):
+            raise ValueError(
+                "Relay transport policy requires a STUN and/or TURN server."
+            )
+
+        self._transport_policy = transport_policy
+
     @property
     def local_candidates(self) -> List[Candidate]:
         """
@@ -363,7 +401,7 @@ class Connection:
 
         # resolve mDNS candidate
         if mdns.is_mdns_hostname(remote_candidate.host):
-            mdns_protocol = await get_or_create_mdns_protocol()
+            mdns_protocol = await get_or_create_mdns_protocol(self)
             remote_addr = await mdns_protocol.resolve(remote_candidate.host)
             if remote_addr is None:
                 self.__log_info(
@@ -457,6 +495,7 @@ class Connection:
         for early_check in self._early_checks:
             self.check_incoming(*early_check)
         self._early_checks = []
+        self._early_checks_done = True
 
         # perform checks
         while True:
@@ -472,30 +511,33 @@ class Connection:
 
         # cancel remaining checks
         for check in self._check_list:
-            if check.handle:
-                check.handle.cancel()
+            if check.task:
+                check.task.cancel()
 
         if res != ICE_COMPLETED:
             raise ConnectionError("ICE negotiation failed")
 
         # start consent freshness tests
-        self._query_consent_handle = asyncio.ensure_future(self.query_consent())
+        self._query_consent_task = asyncio.create_task(self.query_consent())
 
     async def close(self) -> None:
         """
         Close the connection.
         """
         # stop consent freshness tests
-        if self._query_consent_handle and not self._query_consent_handle.done():
-            self._query_consent_handle.cancel()
+        if self._query_consent_task and not self._query_consent_task.done():
+            self._query_consent_task.cancel()
             try:
-                await self._query_consent_handle
+                await self._query_consent_task
             except asyncio.CancelledError:
                 pass
 
         # stop check list
         if self._check_list and not self._check_list_done:
-            await self._check_list_state.put(ICE_FAILED)
+            self._check_list_state.put_nowait(ICE_FAILED)
+
+        # unreference mDNS
+        await unref_mdns_protocol(self)
 
         self._nominated.clear()
         for protocol in self._protocols:
@@ -622,11 +664,13 @@ class Connection:
         return request
 
     def check_complete(self, pair: CandidatePair) -> None:
-        pair.handle = None
+        pair.task = None
 
         if pair.state == CandidatePair.State.SUCCEEDED:
             if pair.nominated:
-                self._nominated[pair.component] = pair
+                # Always choose the one with minimum RTT
+                if pair.component not in self._nominated:
+                    self._nominated[pair.component] = pair
 
                 # 8.1.2.  Updating States
                 #
@@ -646,7 +690,7 @@ class Connection:
             if len(self._nominated) == len(self._components):
                 if not self._check_list_done:
                     self.__log_info("ICE completed")
-                    asyncio.ensure_future(self._check_list_state.put(ICE_COMPLETED))
+                    self._check_list_state.put_nowait(ICE_COMPLETED)
                     self._check_list_done = True
                 return
 
@@ -672,7 +716,7 @@ class Connection:
 
         if not self._check_list_done:
             self.__log_info("ICE failed")
-            asyncio.ensure_future(self._check_list_state.put(ICE_FAILED))
+            self._check_list_state.put_nowait(ICE_FAILED)
             self._check_list_done = True
 
     def check_incoming(
@@ -714,7 +758,7 @@ class Connection:
 
         # triggered check
         if pair.state in [CandidatePair.State.WAITING, CandidatePair.State.FAILED]:
-            pair.handle = asyncio.ensure_future(self.check_start(pair))
+            self.check_start_task(pair)
 
         # 7.2.1.5. Updating the Nominated Flag
         if "USE-CANDIDATE" in message.attributes and not self.ice_controlling:
@@ -728,13 +772,13 @@ class Connection:
         # find the highest-priority pair that is in the waiting state
         for pair in self._check_list:
             if pair.state == CandidatePair.State.WAITING:
-                pair.handle = asyncio.ensure_future(self.check_start(pair))
+                self.check_start_task(pair)
                 return True
 
         # find the highest-priority pair that is in the frozen state
         for pair in self._check_list:
             if pair.state == CandidatePair.State.FROZEN:
-                pair.handle = asyncio.ensure_future(self.check_start(pair))
+                self.check_start_task(pair)
                 return True
 
         # if we expect more candidates, keep going
@@ -804,6 +848,13 @@ class Connection:
         self.check_state(pair, CandidatePair.State.SUCCEEDED)
         self.check_complete(pair)
 
+    def check_start_task(self, pair: CandidatePair) -> None:
+        """
+        Starts a check in a task, unless already started.
+        """
+        if pair.task is None:
+            pair.task = asyncio.create_task(self.check_start(pair))
+
     def check_state(self, pair: CandidatePair, state: CandidatePair.State) -> None:
         """
         Updates the state of a check.
@@ -839,13 +890,17 @@ class Connection:
         for address in addresses:
             # create transport
             try:
-                _, protocol = await loop.create_datagram_endpoint(
+                transport, protocol = await loop.create_datagram_endpoint(
                     lambda: StunProtocol(self), local_addr=(address, 0)
                 )
+                sock = transport.get_extra_info("socket")
+                if sock is not None:
+                    sock.setsockopt(
+                        socket.SOL_SOCKET, socket.SO_RCVBUF, turn.UDP_SOCKET_BUFFER_SIZE
+                    )
             except OSError as exc:
                 self.__log_info("Could not bind to %s - %s", address, exc)
                 continue
-            protocol = cast(StunProtocol, protocol)
             host_protocols.append(protocol)
 
             # add host candidate
@@ -859,7 +914,8 @@ class Connection:
                 port=candidate_address[1],
                 type="host",
             )
-            candidates.append(protocol.local_candidate)
+            if self._transport_policy == TransportPolicy.ALL:
+                candidates.append(protocol.local_candidate)
         self._protocols += host_protocols
 
         # query STUN server for server-reflexive candidates (IPv4 only)
@@ -868,7 +924,7 @@ class Connection:
             for protocol in host_protocols:
                 if ipaddress.ip_address(protocol.local_candidate.host).version == 4:
                     tasks.append(
-                        asyncio.ensure_future(
+                        asyncio.create_task(
                             server_reflexive_candidate(protocol, self.stun_server)
                         )
                     )
@@ -891,7 +947,6 @@ class Connection:
                 ssl=self.turn_ssl,
                 transport=self.turn_transport,
             )
-            protocol = cast(StunProtocol, protocol)
             self._protocols.append(protocol)
 
             # add relayed candidate
@@ -949,7 +1004,7 @@ class Connection:
                     failures += 1
                 if failures >= CONSENT_FAILURES:
                     self.__log_info("Consent to send expired")
-                    self._query_consent_handle = None
+                    self._query_consent_task = None
                     return await self.close()
 
     def data_received(self, data: bytes, component: int) -> None:
@@ -1003,7 +1058,7 @@ class Connection:
         response.add_message_integrity(self.local_password.encode("utf8"))
         protocol.send_stun(response, addr)
 
-        if not self._check_list:
+        if not self._check_list and not self._early_checks_done:
             self._early_checks.append((message, addr, protocol))
         else:
             self.check_incoming(message, addr, protocol)
